------------------------------------------------
--               CT_BarMod                    --
--                                            --
-- Please do not modify or otherwise          --
-- redistribute this without the consent of   --
-- the CTMod Team. Thank you.                 --
------------------------------------------------

--------------------------------------------
-- Initialization

local _G = getfenv(0);
local module = _G.CT_BarMod;

--------------------------------------------
-- Group Management

local group =  {};
local groupMeta = { __index = group };

local groupList = {};
module.groupList = groupList;

local dragOnTop = false;

--------------------------------------------
-- Group frame

local groupFrameTable;

local function groupFrame_OnEnter(self)
	if (not module:getOption("dragHideTooltip")) then
		local text = "|c00FFFFFF" .. self.object.fullName .. "|r\n" ..
			"Left click to drag\n" ..
			"Right click to rotate\n" ..
			"Shift click to reset";
		module:displayTooltip(self, text);
	end
end

local function groupFrame_OnLeave(self)
	module.hideTooltip();
end

local function groupFrame_OnMouseDown(self, button)
	if (button == "LeftButton") then
		-- Hide the tooltip while user is dragging the frame.
		if (GameTooltip:IsOwned(self)) then
			GameTooltip:Hide();
		end
		-- Start moving the frame.
		module:moveMovable(self.movable);
	end
end

local function groupFrame_OnMouseUp(self, button)
	if (button == "LeftButton") then
		if ( IsShiftKeyDown() ) then
			self.object:resetPosition();
		else
			-- Stop moving the frame.
			module:stopMovable(self.movable);

			-- Redisplay the tooltip
			groupFrame_OnEnter(self);
		end
	elseif (button == "RightButton") then
		self.object:rotate();
		-- Update some stuff on the options window
		module.updateGroupWidgets_Orientation(self.object.groupId);
		module.updateGroupWidgets_Columns(self.object.groupId);
	end
end

local function groupFrame_OnLoad(self)
	self:SetBackdropBorderColor(1, 1, 1, 0);
end

local function groupFrameSkeleton()
	-- Frame that is used to anchor the first action button to.
	if ( not groupFrameTable ) then
		groupFrameTable = {
			-- Frame used to allow dragging of the action buttons.
			["button#hidden#i:button#st:LOW"] = {
				"backdrop#tooltip#0:0:0.5:0.85",
				"font#v:GameFontNormalLarge#i:text",
				["onenter"] = groupFrame_OnEnter,
				["onleave"] = groupFrame_OnLeave,
				["onmousedown"] = groupFrame_OnMouseDown,
				["onmouseup"] = groupFrame_OnMouseUp,
				["onload"] = groupFrame_OnLoad,
			}
		};
	end
	return "frame#s:60:30#v:SecureFrameTemplate,SecureHandlerAttributeTemplate", groupFrameTable;  -- width=60, height=30
end

-- Existing CT_BarMod_GroupNN frames are 60 wide.
-- The first button was positioned TOP relative to the frame (ie. centered).
-- Now that we're positioning TOPLEFT relative to the frame,
-- we'll use the following offset when positioning the first button so that
-- existing users' bars will be in the same place.
local firstButtonOffset = 12;

local function checkMouseover(elapsed)
	for groupNum, group in pairs(groupList) do
		if (group.barMouseover) then
			if (group.overlay:IsMouseOver()) then
				-- Mouse is over the frame
				if (group.hideFlag) then
					module:unschedule(group.hideFunc);
					group.hideFlag = false;
				end
				if (not group.mouseover) then
					-- Mouse has just entered the frame
					if (not group.showFlag) then
						module:schedule(group.showTimer, group.showFunc);
						group.showFlag = true;
					end
				end
			else
				-- Mouse is not over the frame
				if (group.showFlag) then
					module:unschedule(group.showFunc);
					group.showFlag = false;
				end
				if (group.mouseover) then
					-- Mouse was previously over the frame
					if (not group.hideFlag) then
						module:schedule(group.hideTimer, group.hideFunc);
						group.hideFlag = true;
					end
				end
			end
		end
	end
end

module.checkMouseover = checkMouseover;

local function setDragOnTop(onTop)
	-- Show group drag frames on top or behind the action buttons.
	if (onTop) then
		dragOnTop = true;
	else
		dragOnTop = false;
	end
	for groupNum, group in pairs(groupList) do
		local button = group.frame.button;
		group:updateButtonPosition();
		if (dragOnTop) then
			button:SetFrameLevel(button.originalLevel + 2);
		else
			button:SetFrameLevel(button.originalLevel)
		end
		group:updateOverlayPosition();
	end
end

module.setDragOnTop = setDragOnTop;


--------------------------
-- Group Class

local function getGroup(groupId)
	local groupNum = module.GroupIdToNum(groupId);
	return groupList[groupNum] or group:new(groupId);
end


module.updateActionBar = function(pos, orientation, scale, opacity, spacing, hide)
	-- Update the action bar.
	-- This was added so that CT_BottomBar could update the action bar.
	-- This is intended to be run before CT_BarMod's option window gets opened.
	-- This function does not update option window widgets.

	local groupId = module.actionBarId;
	local obj = getGroup(groupId);
	local frame = obj.frame;

	if (type(pos) == "table") then
		-- Set the position of the action bar.
		frame:ClearAllPoints();
		frame:SetPoint(pos[1] or "TOPLEFT", UIParent, pos[3] or "TOPLEFT", pos[4] or 0, pos[5] or 0);
		module:stopMovable(obj.movable);

	else
		-- Reset action bar to default position.
		obj:resetPosition();
	end

	local optName, value;

	optName = "orientation";
	value = orientation;
	module:setOption(optName .. groupId, value, true);
	obj:update(optName, value);

	optName = "barScale";
	value = scale;
	module:setOption(optName .. groupId, value, true);
	obj:update(optName, value);

	optName = "barOpacity";
	value = opacity;
	module:setOption(optName .. groupId, value, true);
	obj:update(optName, value);

	optName = "barSpacing";
	value = spacing;
	module:setOption(optName .. groupId, value, true);
	obj:update(optName, value);

	optName = "showGroup";
	value = not hide;
	module:setOption(optName .. groupId, value, true);
end

module.resetActionBarPosition = function()
	-- Reset the position of the action bar.
	-- This was added so that CT_BottomBar could reset the position of the action bar.
	local obj = getGroup(module.actionBarId);
	if (obj) then
		obj:resetPosition();
	end
end

module.resetAllBarPositions = function()
	-- Reset the position of the action bar.
	-- This was added so that CT_BottomBar could reset the position of the bars.
	for groupNum, obj in pairs(groupList) do
		obj:resetPosition();
	end
end

local setActionPage_unsecure = function(self, page)
	-- Parameters: self, stateid, newstate
	-- self == group frame
	-- page == action bar page number (1 to 11).

	local base = (page - 1) * 12;
	local count = 1;
	local actionId;
	local button = self:GetAttribute("child1");
	while (button) do
		actionId = base + count;
		button.action = actionId;
		button.object.actionId = actionId;
		button.object:update();
		button.object:updateRange();
		count = count + 1;
		button = self:GetAttribute("child" .. count);
	end

	-- Blizzard doesn't update the page number font string on the
	-- action bar arrows when you get into a vehicle, even though
	-- the game has changed the action bar page to 1 (GetActionBarPage() == 1).
	MainMenuBarPageNumber:SetText(GetActionBarPage());

	-- Update our key bindings list if the window is visible.
	module.keybindings_buttonsUpdateList();
end

local setActionPage_secure = [=[
	-- Parameters: self, pagenum
	-- self == group frame
	-- (1, ...) == action bar page number (1 to 11).

	local page = floor(tonumber(select(1, ...)) or 1);
	if (page < 1 or page > 11) then
		page = 1;
	end

	local usePage = page;

	local base = (usePage - 1) * 12;
	local count = 1;
	local actionId;
	local button = self:GetFrameRef("child1");
	while (button) do
		actionId = base + count;
		button:SetAttribute("action", actionId);
		if (actionId == 132) then -- module.controlCancel
			-- Since this action id isn't used for anything, we're
			-- going to use it to display a button that can be used
			-- to cancel control.
			-- (refer to a longer description in CT_BarMod_Use.lua)
			-- There are a few other places in this addon that tests
			-- for this action id number.

			-- Set the "type" attribute based on whether or not the
			-- cancel possess button is enabled. That button when clicked
			-- will either exit a vehicle or cancel a possession spell.
			--
			-- If the game does not provide possession information then
			-- it cannot be used otherwise the game will try to cancel
			-- a nil buff causing an error.
			--
			-- We're using secure wrapper around Blizzard's PossessButton2's
			-- OnShow and OnHide scripts to tell us when the possess information
			-- is available and the possess button can be clicked. This is
			-- necessary because GetPossessInfo() is not callable from
			-- secure snippets.
			--
			-- For similar "type" attribute code see also:
			-- 1) setActionPage_Secure in CT_BarMod_Use.lua
			-- 2) secureFrame_OnAttributeChanged in CT_BarMod.lua
			-- 3) initUpdateButtonType() in CT_BarMod.lua
			--
			if (UnitHasVehicleUI("player")) then
				button:SetAttribute("type", "click");
			else
				local secureFrame = button:GetFrameRef("SecureFrame");
				if (secureFrame:GetAttribute("showcancel")) then
					button:SetAttribute("type", "click");
				else
					-- This is to prevent Blizzard's code from causing an error
					-- if the user clicks the button when there is no possess info
					-- available.
					button:SetAttribute("type", nil);
				end
			end

			-- We can't use a secure frame reference to a button since the
			-- frame reference does not have a "Click" method that the
			-- SecureTemplates.lua routine wants to use when our action button
			-- gets clicked.
			-- The button needs to be assigned to the attribute while in unsecure
			-- code (see CT_BarModUse.lua useButton:constructor()).
			--button:SetAttribute("clickbutton", button:GetFrameRef("PossessButton2"));
		else
			button:SetAttribute("type", "action");
		end

		do
			-- We need to update the show/hide state of the button when the page changes.
			-- If a user has the "hide empty button grid" option enabled, and they have an
			-- empty button on page 2, and they switch to page 1 while in combat, then the
			-- unsecure code can't unhide the button that was hidden while page 2 was
			-- displayed on the bar, so have to do it here in secure code.

			local show;

			-- If we are showing this button (ie. it is not being forced hidden)...
			if (button:GetAttribute("showbutton")) then

				-- If a show grid event is not currently active...
				if (button:GetAttribute("gridevent") == 0) then

					-- If the button has an action...
					local actionId = button:GetAttribute("action");
					if ( HasAction(actionId) or actionId == 132 ) then -- 132 == module.controlCancel
						show = true;
					else
						-- The button has no action.
						-- If we want to show empty buttons...
						if (button:GetAttribute("gridshow") > 0) then
							show = true;
						end
					end
				else
					-- There is a show grid event that is currently active.
					-- The user is probably dragging an action button.
					-- Show all buttons, empty or not.
					show = true;
				end
			end

			if (show) then
				button:Show();
			else
				button:Hide();
			end
		end

		count = count + 1;
		button = self:GetFrameRef("child" .. count);
	end

	self:SetAttribute("currentpage", usePage);

	-- Perform visual updates of the buttons
	self:CallMethod("setActionPage", usePage);
]=];

local group_OnAttributeChanged_secure = [=[
	-- Parameters: self, name, value
	if (name == "state-barpage") then
		-- Bar state
		self:RunAttribute("setActionPage", value);
	
	end
]=];

local firstTime;
function group:new(groupId)
	local group = { };
	local frame = module:getFrame(groupFrameSkeleton, UIParent, "CT_BarMod_Group" .. groupId);  -- skeleton, parent, name

	local frameLevel = 10;  -- high enough to get above the ActionButton's and BonusActionButton's
	frame:SetFrameLevel(frameLevel);

--	local backdrop = {
--		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
--		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
--		tile = true,
--		tileSize = 16,
--		edgeSize = 16,
--		insets = { left = 5, right = 5, top = 5, bottom = 5 },
--	};
--	frame:SetBackdrop(backdrop);
--	frame:SetBackdropColor(1, 1, 1, 0.85);

	frame.button:SetParent(UIParent);
	frame.button:SetHitRectInsets(5, 5, 5, 5);
	frame.button:SetBackdropColor(0, 0, 0.5, 0.85);

	frame.button:SetFrameLevel(frameLevel);
	frame.button.originalLevel = frame.button:GetFrameLevel();
	if (dragOnTop) then
		frame.button:SetFrameLevel(frame.button:GetFrameLevel()+2);
	end

	frame.object = group;

	local movable = "GROUP"..groupId;
	setmetatable(group, groupMeta);
	
	group.orientation = module:getOption("orientation"..groupId);
	group.frame = frame;
	group.groupId = groupId;
	group.movable = movable;
	group.num = module.GroupIdToNum(groupId);

	-- .fullName is used in drag frame tooltip, and in keybindings section.
	-- .longName is displayed on bar.
	-- .shortName is displayed on bar.
	if (group.groupId == module.actionBarId) then
		group.fullName = "Bar 12 (Action bar)";
		group.longName = "Action bar";
		group.shortName = "Act";
	elseif (group.groupId == module.controlBarId) then
		group.fullName = "Bar 11 (Control bar)";
		group.longName = "Control bar";
		group.shortName = "Ctl";
	else
		group.fullName = "Bar " .. group.num;
		group.longName = "Bar " .. group.num;
		group.shortName = "B" .. group.num;
	end

	group.barColumns = 1; -- Desired number of columns (across) or rows (down). Init to 1.
	group.numColumns = 1; -- Actual number of columns (across) or rows (down). Init to 1.
	group.numRows = 1;  -- Actual number of rows (across) or columns (down). Init to 1.
	group.barNumToShow = 12; -- Number of buttons to show. Remaining buttons are kept hidden. Init to 12.

	-- Do the following option tests and sets before we add the group to the groupList table.

	if (module:getOption("orientation" .. groupId) == nil) then
		-- This character does not have a value assigned to this option yet.
		-- If this is not one of the original 5 bars (groupIds 1 through 5),
		-- then initially hide this group so that it does not appear in the middle
		-- of the user's screen. They can toggle it on via the CT_BarMod options.
		if (groupId >= 6) then
			module:setOption("showGroup" .. groupId, false, true);
		end
	end

	if (module:getOption("barHideInVehicle" .. groupId) == nil) then
		-- This character does not have a value assigned to this option yet.
		-- This option was added in 3.304 to replace an existing vehicleHideOther option.
		-- Initialize the new option to the same setting as the old option.
		local hide = module:getOption("vehicleHideOther"); -- default was 1, 1=hide, 2=hide, 3=hide, 4=show
		if (hide ~= nil) then
			if (hide == 4) then
				hide = false;
			else
				hide = 1;
			end
		end
		-- If the bar is not enabled...
		if (not module:getOption("showGroup" .. groupId)) then
			if (group.groupId == module.actionBarId) then
				-- Don't hide main action bar when in a vehicle
				hide = false;
			elseif (group.groupId == module.controlBarId) then
				-- Don't hide control bar when in a vehicle
				hide = false;
			end
		end
		module:setOption("barHideInVehicle" .. groupId, hide, true);
	end

	if (module:getOption("stdPositions") == nil) then
		-- If this option is nil, then this is a new character,
		-- or an existing character's first time for this option.
		if (module:getOption("orientation1") == nil) then
			-- New character
			local newPositions = module:getOption("newPositions");
			if (newPositions == nil) then
				-- Default to using the standard bar positions.
				newPositions = true;
			end
			module:setOption("stdPositions", newPositions, true);
		else
			-- Existing character's first time for this option.
			-- Default to the original CT_BarMod positions value (false) since
			-- those would be the bar positions that they have been using.
			module:setOption("stdPositions", false, true);
		end
		firstTime = true;
	end

	group:position(group.orientation, module:getOption("stdPositions"));

	groupList[group.num] = group;
	
	local button = frame.button;
	button.movable = movable;
	button.object = group;
	module:registerMovable(movable, frame);

	group:setClamped(not not module:getOption("clampFrames"));  -- Set after the regiserMovable() call.

	-- Create an invisible frame and overlay it on the frame used to drag the bar group.
	-- This will allow us to detect OnEnter and OnLeave events when the drag frame is not visible.
	local f = CreateFrame("Frame", "CT_BarMod_Group" .. groupId .. "Frame", frame);

	group.overlay = f;
	group:updateOverlayPosition();

	f:SetHitRectInsets(10, 10, 10, 10);
	f:Show();
	f.groupId = groupId;

	group.mouseover = false;
	group.hideFlag = false;
	group.showFlag = false;
	group.hideTimer = 0;
	group.showTimer = 0;

	group.hideFunc = function()
		group.hideFlag = false;
		group.mouseover = false;
		group:updateOpacity();
	end

	group.showFunc = function()
		group.showFlag = false;
		group.mouseover = true;
		group:updateOpacity();
	end


	-- Set the attribute that will tell the game which secure function to call.
	group.frame:SetAttribute("_onattributechanged", group_OnAttributeChanged_secure);

	-- A secure snippet to change the action page.
	group.frame:SetAttribute("setActionPage", setActionPage_secure);

	-- An unsecure method for the secure function to call.
	group.frame.setActionPage = setActionPage_unsecure;

	-- Add a reference to this group, to the secure ActionBarButton we're using to deal
	-- with ACTIONBAR_SHOWGRID and ACTIONBAR_HIDEGRID events.
	SecureHandlerSetFrameRef(CT_BarMod_ActionBarButton, "group" .. group.num, group.frame);
	CT_BarMod_ActionBarButton:SetAttribute("group" .. group.num, group.frame);

	-- Add a reference to this group, to our secure frame.
	SecureHandlerSetFrameRef(CT_BarMod_SecureFrame, "group" .. group.num, group.frame);
	CT_BarMod_SecureFrame:SetAttribute("group" .. group.num, group.frame);

	return group;
end

function group:updateOpacity()
	self:update("barFaded", module:getOption("barFaded" .. self.groupId) or 0);
	self:update("barOpacity", module:getOption("barOpacity" .. self.groupId) or 1);
end

function group:updateOverlayPosition()
	local f = self.overlay;
	local frame = self.frame;
	f:ClearAllPoints();
	if (dragOnTop) then
		f:SetPoint("TOPLEFT", frame.button, "TOPLEFT", 8, -8);
		f:SetPoint("BOTTOMRIGHT", frame.button, "BOTTOMRIGHT", -8, 8);
	else
		f:SetPoint("TOPLEFT", frame.button, "TOPLEFT", 11, -11);
		f:SetPoint("BOTTOMRIGHT", frame.button, "BOTTOMRIGHT", -11, 11);
	end
	f:SetFrameLevel(f:GetParent():GetFrameLevel());
end

local defaultPositions = {};

-- These are the original bar positions used by CT_BarMod.
--   Bar  4 = LE = Left
--   Bar  2 = BL = Bottom left
--   Bar  3 = BR = Bottom right
--   Bar  6 = IR = Inside right
--   Bar  5 = OR = Outside right
--   Bar  7 = BC = Bottom center (lowest)
--   Bar  8 = LC = Lower center
--   Bar  9 = UC = Upper center
--   Bar 10 = TC = Top center
--   Bar  1 = AC = Above center
--   Bar 11 = OC = Over center
--   Bar 12 = HC = High center (highest)
--
-- Secondary index is the bar (group) id.
defaultPositions[1] = {
	[10] = {"BOTTOMLEFT",  "BOTTOM",      -260, 560, "ACROSS", "AC"},  -- Bar  1, Above center
	[1]  = {"BOTTOMLEFT",  "BOTTOM",      -516,  97, "ACROSS", "BL"},  -- Bar  2, Bottom left
	[2]  = {"BOTTOMLEFT",  "BOTTOM",        -6,  97, "ACROSS", "BR"},  -- Bar  3, Bottom right
	[3]  = {"BOTTOMLEFT",  "TOPLEFT",      -10, -85, "DOWN",   "LE"},  -- Bar  4, Left
	[4]  = {"BOTTOMRIGHT", "BOTTOMRIGHT",   12, 603, "DOWN",   "OR"},  -- Bar  5, Outside right
	[5]  = {"BOTTOMRIGHT", "BOTTOMRIGHT",  -31, 603, "DOWN",   "IR"},  -- Bar  6, Inside right
	[6]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 300, "ACROSS", "BC"},  -- Bar  7, Bottom center
	[7]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 365, "ACROSS", "LC"},  -- Bar  8, Lower center
	[8]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 430, "ACROSS", "UC"},  -- Bar  9, Upper center
	[9]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 495, "ACROSS", "TC"},  -- Bar 10, Top center
	[11] = {"BOTTOMLEFT",  "BOTTOM",      -260, 625, "ACROSS", "OC"},  -- Bar 11, Over center
	[12] = {"BOTTOMLEFT",  "BOTTOM",      -260, 690, "ACROSS", "HC"},  -- Bar 12, High center
};

-- The following standard (Blizzard) bar positions reorganize the groups
-- so that the bar contents match the bars that Blizzard places in the
-- same screen locations. This makes it easier to switch between using CT_BarMod
-- and the default Blizzard bars since the icons on the bars will be the same.
--   Bar  2 = LE = Left
--   Bar  6 = BL = Bottom left
--   Bar  5 = BR = Bottom right
--   Bar  4 = IR = Inside right
--   Bar  3 = OR = Outside right
--   Bar  7 = BC = Bottom center (lowest)
--   Bar  8 = LC = Lower center
--   Bar  9 = UC = Upper center
--   Bar 10 = TC = Top center
--   Bar  1 = AC = Above center
--   Bar 11 = OC = Over center
--   Bar 12 = HC = High center (highest)
--
-- Secondary index is the bar (group) id.
defaultPositions[2] = {
	[10] = {"BOTTOMLEFT",  "BOTTOM",      -260, 560, "ACROSS", "AC"},  -- Bar  1, Above center
	[1]  = {"BOTTOMLEFT",  "TOPLEFT",      -10, -85, "DOWN",   "LE"},  -- Bar  2, Left
	[2]  = {"BOTTOMRIGHT", "BOTTOMRIGHT",   12, 603, "DOWN",   "OR"},  -- Bar  3, Outside right
	[3]  = {"BOTTOMRIGHT", "BOTTOMRIGHT",  -31, 603, "DOWN",   "IR"},  -- Bar  4, Inside right
	[4]  = {"BOTTOMLEFT",  "BOTTOM",        -6,  97, "ACROSS", "BR"},  -- Bar  5, Bottom right
	[5]  = {"BOTTOMLEFT",  "BOTTOM",      -516,  97, "ACROSS", "BL"},  -- Bar  6, Bottom left
	[6]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 300, "ACROSS", "BC"},  -- Bar  7, Bottom center
	[7]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 365, "ACROSS", "LC"},  -- Bar  8, Lower center
	[8]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 430, "ACROSS", "UC"},  -- Bar  9, Upper center
	[9]  = {"BOTTOMLEFT",  "BOTTOM",      -260, 495, "ACROSS", "TC"},  -- Bar 10, Top center
	[11] = {"BOTTOMLEFT",  "BOTTOM",      -260, 625, "ACROSS", "OC"},  -- Bar 11, Over center
	[12] = {"BOTTOMLEFT",  "BOTTOM",      -260, 690, "ACROSS", "HC"},  -- Bar 12, High center
};

-- This holds the position for the main action bar when we detect that CT_BottomBar has disabled
-- the game's main action bar.
--
-- Secondary index is the bar (group) id.
defaultPositions[3] = {
	[12] = {"BOTTOMLEFT",  "BOTTOM",      -516,  43, "ACROSS", "AB"},  -- Bar 12, Action Bar position
};

function group:position(orientation, stdPositions)
	if (InCombatLockdown()) then
		return;
	end
	local groupId = self.groupId;
	local frame = self.frame;
	local pos;
	local yoffset;
	if (stdPositions) then
		pos = defaultPositions[2][groupId];
	else
		pos = defaultPositions[1][groupId];
	end
	if (groupId == module.actionBarId) then
		-- If we detect that CT_BottomBar has disabled the default main action bar
		-- then use this position for the action bar. It will place it in the same
		-- spot as the default UI's main action bar.
		if (CT_BottomBar and CT_BottomBar.actionBarDisabled) then
			pos = defaultPositions[3][groupId];
		end
	end
	if (not pos) then
		pos = defaultPositions[2][module.controlBarId];
	end
	yoffset = pos[4];
	if (pos[6] == "BL" or pos[6] == "BR" or pos[6] == "OR" or pos[6] == "IR") then
		if (firstTime and CT_BottomBar) then
			yoffset = yoffset + 9;
		else
			if (MainMenuBarMaxLevelBar and MainMenuBarMaxLevelBar:IsShown()) then
				-- Max level bar only
				yoffset = yoffset - 5;
			else
				if (ReputationWatchBar and ReputationWatchBar:IsShown()) then
					if (MainMenuExpBar and MainMenuExpBar:IsShown()) then
						-- Rep and Exp bars
						yoffset = yoffset + 9;
					-- else
						-- Rep bar only
						-- yoffset = yoffset + 0;
					end
				-- else
					-- Exp bar only
					-- yoffset = yoffset + 0;
				end
			end
		end
	end
	if (TitanMovable_GetPanelYOffset and TITAN_PANEL_PLACE_BOTTOM and TitanPanelGetVar) then
		if (pos[6] == "BL" or pos[6] == "BR") then
			yoffset = yoffset + (tonumber( TitanMovable_GetPanelYOffset(TITAN_PANEL_PLACE_BOTTOM, TitanPanelGetVar("BothBars")) ) or 0);
		end
	end
	frame:ClearAllPoints();
	frame:SetPoint(pos[1], UIParent, pos[2], pos[3], yoffset);
	self:rotate(orientation or pos[5]);
end

function group:resetPosition()
	-- Reset position of the group
	self:position(nil, module:getOption("stdPositions"));
	module:stopMovable(self.movable);
end

function group:positionButtons()
	local objects = self.objects;
	if ( not objects ) then
		return;
	end
	
	if (InCombatLockdown()) then
		return;
	end

	local frame = self.frame;
	local buttons = self:getNumButtonsToShow();
	local button;
	local offset = self.spacing or 6;
	if ( self.orientation == "DOWN" ) then
		-- DOWN (top to bottom)
		local row, rows, column, columns;
		rows = self.numRows;
		columns = self.numColumns;
		row = 1;
		column = 1;
		for key, object in ipairs(objects) do
			button = object.button;
			button:ClearAllPoints();
			if (key > buttons or row > rows) then
				button:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
				object:updateVisibility();
			else
				if (key == 1) then
					button:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", firstButtonOffset, -4);
				elseif (column == 1) then
					button:SetPoint("TOPLEFT", objects[key - columns].button, "TOPRIGHT", offset, 0);
				else
					button:SetPoint("TOPLEFT", objects[key-1].button, "BOTTOMLEFT", 0, -offset);
				end
				object:updateVisibility();
			end
			column = column + 1;
			if (column > columns) then
				column = 1;
				row = row + 1;
			end
		end
	else
		-- ACROSS (left to right)
		local row, rows, column, columns;
		rows = self.numRows;
		columns = self.numColumns;
		row = 1;
		column = 1;
		for key, object in ipairs(objects) do
			button = object.button;
			button:ClearAllPoints();
			if (key > buttons or row > rows) then
				button:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
				object:updateVisibility();
			else
				if (key == 1) then
					button:SetPoint("TOPLEFT", frame, "BOTTOMLEFT", firstButtonOffset, -4);
				elseif (column == 1) then
					button:SetPoint("TOPLEFT", objects[key - columns].button, "BOTTOMLEFT", 0, -offset);
				else
					button:SetPoint("TOPLEFT", objects[key-1].button, "TOPRIGHT", offset, 0);
				end
				object:updateVisibility();
			end
			column = column + 1;
			if (column > columns) then
				column = 1;
				row = row + 1;
			end
		end
	end
	
	self:updateButtonPosition();
end

function group:updateClampRectInsets()
	-- Adjust the clamping insets

	-- self.frame (which is the movable frame) is
	-- a small rectangle that the first action button is
	-- anchored to. We need to adjust the clamping
	-- insets so that they encompass all of the
	-- action buttons.

	local frame = self.frame;
	local button = frame.button;

	local left, right, top, bottom;

	local frameLeft = frame:GetLeft() or 0;
	local frameRight = frame:GetRight() or 0;
	local frameTop = frame:GetTop() or 0;
	local frameBottom = frame:GetBottom() or 0;

	local buttonLeft = button:GetLeft() or 0;
	local buttonRight = button:GetRight() or 0;
	local buttonTop = button:GetTop() or 0;
	local buttonBottom = button:GetBottom() or 0;
	local buttonHeight = button:GetHeight() or 0;

	local edgeSize = 11;
	if (dragOnTop) then
		edgeSize = 8;
	end

	if (frameLeft < buttonLeft) then
		left = (buttonLeft - frameLeft) + edgeSize;
	else
		left = edgeSize - (frameLeft - buttonLeft);
	end

	if (frameRight > buttonRight) then
		right = -((frameRight - buttonRight) + edgeSize);
	else
		right = (buttonRight - frameRight) - edgeSize;
	end

	top = -((frameTop - buttonTop) + edgeSize);

	if (frameBottom > buttonTop) then
		bottom = -(buttonHeight - edgeSize + (frameBottom - buttonTop));
	else
		bottom = -(buttonHeight - edgeSize - (buttonTop - frameBottom));
	end

	frame:SetClampRectInsets(left, right, top, bottom);
end

function group:setClamped(clamp)
	-- Set clamped state of the frame
	self.frame:SetClampedToScreen(clamp);
	self:updateClampRectInsets();
end

function group:getNumButtonsToShow()
	-- Determine how many of the buttons to show.
	-- The first n buttons will be shown, and the last n buttons will be hidden.
	local buttons;
	local objects = self.objects;
	if (not objects) then
		buttons = 0;
	else
		buttons = #objects;
		if (buttons > self.barNumToShow) then
			buttons = self.barNumToShow;
		end
	end
	return buttons;
end

function group:calculateDimensions()
	-- Determine how many rows and columns are needed.
	--
	-- The user selects how many columns/rows they want to see, and this routine
	-- will determine how many rows/columns will be needed.
	--
	-- The actual interpretation of the number chosen by the user depends on the
	-- bar orientation being used.
	--
	-- In "ACROSS" orientation, the columns var represents columns, and the rows var represents rows.
	-- In "DOWN" orientation, the columns var represents rows, and the rows var represents columns.
	--
	local objects = self.objects;
	local buttons = self:getNumButtonsToShow();

	local columns = self.barColumns;  -- desired number of columns
	if (columns > buttons) then
		columns = buttons;
	end
	if (columns <= 0) then
		columns = 1;
	end

	local rows = floor((buttons - 1) / columns) + 1;
	if (rows > buttons) then
		rows = buttons;
	end
	if (rows <= 0) then
		rows = 1;
	end

	-- Actual number of rows and columns.
	self.numRows = rows;
	self.numColumns = columns;
end

function group:addObject(object)
	local objects = self.objects;
	if (not objects) then
		objects = {};
		self.objects = objects;
	end

	local button = object.button;
	local frame = self.frame;
	
	-- Add the object to the list.
	local lastObject = objects[#objects];  -- Last object in list before we add this one.
	tinsert(objects, object);

	-- Determine how many rows and columns are needed.
	self:calculateDimensions();

	-- Create a frame reference and an attribute reference to the button.
	local buttonCount = #objects;
	SecureHandlerSetFrameRef(frame, "child" .. buttonCount, button);
	frame:SetAttribute("child" .. buttonCount, button);

	-- Prepare to position the object
	if (InCombatLockdown()) then
		return;
	end

	button:SetParent(frame);
	button:ClearAllPoints();

	if ( not lastObject ) then
		-- This object is the only one in the group so far.
		button:SetPoint("TOPLEFT", self.frame, "BOTTOMLEFT", firstButtonOffset, -4);
		return;
	end
	
	local buttons = self:getNumButtonsToShow();
	if (#objects > buttons) then
		button:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
		object:updateVisibility();
	else
		-- Calculate row and column of the object.
		local columns = self.numColumns;
		local row = floor((buttons - 1) / columns) + 1;
		local column = buttons - ((row - 1) * columns);

		-- Attach the object to the group.
		local lastButton = lastObject.button;
		local offset = self.spacing or 6;

		if ( self.orientation == "DOWN" ) then
			if (column == 1) then
				button:SetPoint("TOPLEFT", objects[buttons - columns].button, "TOPRIGHT", offset, 0);
			else
				button:SetPoint("TOPLEFT", lastButton, "BOTTOMLEFT", 0, -offset);
			end
		else
			if (column == 1) then
				button:SetPoint("TOPLEFT", objects[buttons - columns].button, "BOTTOMLEFT", 0, -offset);
			else
				button:SetPoint("TOPLEFT", lastButton, "TOPRIGHT", offset, 0);
			end
		end
		object:updateVisibility();
	end
	button:SetScale(self.scale or 1);
	button:SetAlpha(self.opacity or 1);
	
	self:updateButtonPosition();
end

function group:updateButtonPosition()
	-- Update position of the button that allows the user to drag the group.
	local button = self.frame.button;
	local objects = self.objects;
	if ( not objects ) then
		return;
	end
	
	if (InCombatLockdown()) then
		return;
	end

	local buttons = self:getNumButtonsToShow();

	-- Determine which buttons will be used to anchor
	-- the top right and bottom left corners of the
	-- drag frame.
	local tr, bl;
	tr = self.numColumns;
	if (buttons < tr) then
		tr = buttons;
	end
	bl = (self.numRows - 1) * self.numColumns + 1;
	if (buttons < bl) then
		-- Calculate number of first object on same row as the last object.
		bl = floor((buttons - 1) / self.numColumns) * self.numColumns + 1;
	end
	if ( self.orientation == "DOWN" ) then
		local temp = tr;
		tr = bl;
		bl = temp;
	end

	-- Anchor three corners of the drag frame to buttons.
	local offset;
	if (dragOnTop) then
		offset = 8;
	else
		offset = 11;
	end
	button:ClearAllPoints();
	button:SetPoint("TOPLEFT", objects[1].button, -offset, offset);
	button:SetPoint("TOPRIGHT", objects[tr].button, offset, offset);
	button:SetPoint("BOTTOMLEFT", objects[bl].button, -offset, -offset);
	
	-- Display the name of the group.
	local text = button.text;
	text:ClearAllPoints();
	if (dragOnTop) then
		text:SetPoint("CENTER", button, "CENTER", 0, 0);
	else
		text:SetPoint("BOTTOM", button, "TOP", 0, -5);
	end
	if ( self.orientation == "ACROSS" ) then
		if (self.numColumns == 1) then
			text:SetText(self.shortName);
		else
			text:SetText(self.longName);
		end
	else
		if (self.numRows == 1) then
			text:SetText(self.shortName);
		else
			text:SetText(self.longName);
		end
	end

	self:updateClampRectInsets();
end

function group:rotate(force)
	if ( force ) then
		self.orientation = force;
	else
		if ( self.orientation == "DOWN" ) then
			self.orientation = "ACROSS";
		else
			self.orientation = "DOWN";
		end
	end
	module:setOption("orientation"..self.groupId, self.orientation, true);
	self:positionButtons();
end

function group:show()
	if (InCombatLockdown()) then
		return;
	end
--	self.frame:Show();
end

function group:hide()
	if (InCombatLockdown()) then
		return;
	end
--	self.frame:Hide();
end

function group:toggleHeader(show)
	if ( show ) then
		self.frame.button:Show();
	else
		self.frame.button:Hide();
	end
end

function group:update(optName, value)
	if ( optName == "barScale" ) then
		self.scale = value;
		if (InCombatLockdown()) then
			return;
		end
		local objects = self.objects;
		if ( objects ) then
			for key, object in ipairs(objects) do
				object.button:SetScale(value);
			end
		end
		self:updateClampRectInsets();

	elseif ( optName == "barMouseover" ) then
		self.barMouseover = value;

	elseif ( optName == "barFaded" ) then
		self.barFaded = value;

	elseif ( optName == "barOpacity" ) then
		-- Init "barMouseover" and/or "barFaded" before "barOpacity".
		self.opacity = value;
		if (self.barMouseover) then
--			-- If the group headers are not visible (the CT_BarMod/CT_BottomBar options window are closed)...
--			if (not module.showingHeaders) then
				-- If mouse is not over this bar group...
				if (not self.mouseover) then
					value = (self.barFaded or 0);  -- Fade buttons by changing alpha.
				end
--			end
		end
		local objects = self.objects;
		if ( objects ) then
			for key, object in ipairs(objects) do
				object.alphaCurrent = value;
				object:updateOpacity();
			end
		end

	elseif ( optName == "barSpacing" ) then
		self.spacing = value;
		self:positionButtons();

	elseif ( optName == "showGroup" ) then
		if (InCombatLockdown()) then
			return;
		end
		if ( value ) then
			self:show();
		else
			self:hide();
		end

	elseif ( optName == "orientation" ) then
		self:rotate(value);  -- value is "ACROSS" or "DOWN"

	elseif ( optName == "barNumToShow" ) then
		-- Number of buttons to show, starting from the first one (topmost, or leftmost).
		value = value or 12;
		self.barNumToShow = value;

		local objects = self.objects;
		if ( objects ) then
			-- Flag each button for showing or hiding.
			local show;
			for key, object in ipairs(objects) do
				if (key > value) then
					show = false;
				else
					show = true;
				end
				object.showButton = show;
				if (not InCombatLockdown()) then
					object.button:SetAttribute("showbutton", show);
				else
					module.needSetAttributes = true;
				end
			end
		end

		-- Any time we change the number of buttons to show on the bar ("barNumToShow")
		-- we also want to change the upper limit of the slider that controls the desired
		-- number of columns/rows ("barColumns"). The user should not be allowed to select
		-- a "barColumns" value that exceeds the number of buttons we want to show.
		--
		-- The upper limit of the slider will get changed during the function that updates
		-- the "barColumns" widgets on the options screen. However, that function won't get
		-- called until after we return from this current function.
		--
		-- When the upper limit of that slider is changed, it will force the "barColumns" option
		-- to be changed if the current value is greater than the new limit.
		--
		-- Before we return from this function, we want to adjust "self.barColumns" if
		-- necessary, so that the new value will be used in the functions we are going to
		-- call before returning. This will be the same test that occurs when the slider's
		-- max value is eventually changed, but here we're not changing the actual "barColumns"
		-- option.

		if (self.barColumns > self.barNumToShow) then
			-- Adjust the group's copy of the "barColumns" option so that it will get
			-- used in the functions we're about to call before returning.
			self.barColumns = self.barNumToShow;
		end

		self:calculateDimensions();
		self:positionButtons();

	elseif ( optName == "barColumns" ) then
		-- Change the number of columns (or rows) that the user would
		-- like to see. The actual number of columns and rows will
		-- then be calculated.
		self.barColumns = value or 12;
		self:calculateDimensions();
		self:positionButtons();

	end
end

--------------------------------------------
-- ACTIONBAR_SHOWGRID and ACTIONBAR_HIDEGRID event handling

-- We want to be able to handle the ACTIONBAR_SHOWGRID and ACTIONBAR_HIDEGRID events in secure code.
-- To do this we'll create an ActionBarButton that inherits from ActionBarButtonTemplate in ActionBarFrame.xml.
-- Blizzard's OnEvent script for these events call functions which change an attribute.
-- We can use a secure wrapper around the button's OnAttributeChanged script so that we can
-- get into secure code in order to show/hide our buttons in response to these events.

local actionButton_gridevent_Unsecure = function(self, showgrid)
	-- Parameters: self, name, value
	-- self == CT_BarMod_ActionBarButton
	-- showgrid == true if ACTIONBAR_SHOWGRID event, else false if ACTIONBAR_HIDEGRID event.

	-- Process all groups.
	local groupNum = 1;
	local group = self:GetAttribute("group1");
	while (group) do
		-- Process all buttons associated with the group.
		local count = 1;
		local button = group:GetAttribute("child1");
		while (button) do

			if (showgrid) then
				button.normalTexture:SetVertexColor(1, 1, 1, 0.5);
			else
				button.object:updateUsable();
			end

			-- Next button
			count = count + 1;
			button = group:GetAttribute("child" .. count);
		end
		-- Next group
		groupNum = groupNum + 1;
		group = self:GetAttribute("group" .. groupNum);
	end
end

local actionButton_OnAttributeChanged_Pre_Secure = [=[
	-- Pre wrap of a securely wrapped OnAttributeChanged script for a button that
	-- inherits from ActionBarButtonTemplate in ActionBarFrame.xml
	--
	-- This is to allow us to handle the "ACTIONBAR_SHOWGRID" and "ACTIONBAR_HIDEGRID"
	-- events via secure code.
	--
	-- Params: self, name, value

	if (name == "showgrid") then
		-- We need to show or hide the empty buttons where applicable.

		-- If new value is greater than previous value, then this was an ACTIONBAR_SHOWGRID event,
		-- else it was an ACTIONBAR_HIDEGRID event.
		local showgrid;
		if (not previousValue) then
			previousValue = 0;
		end
		if (value > previousValue) then
			showgrid = true;
		end
		previousValue = value;

		-- Process all groups.
		local groupNum = 1;
		local group = self:GetFrameRef("group1");
		while (group) do
			-- Process all buttons associated with the group.
			local count = 1;
			local button = group:GetFrameRef("child1");
			while (button) do

				-- For similar visibility code see also:
				-- 1) actionButton_OnAttributeChanged_Pre_Secure in CT_BarMod_Groups.lua
				-- 2) useButton:updateVisibility() in CT_BarMod_Use.lua
				-- 3) SecureHandlerWrapScript(button, "PostClick", ... in CT_BarMod_Use.lua
				-- 4) secureFrame_OnAttributeChanged in CT_BarMod.lua

				-- Show or hide the button
				local show;
				if (showgrid) then
					-- Increase the gridevent count
					button:SetAttribute("gridevent", button:GetAttribute("gridevent") + 1);

					-- If we are showing this button (ie. it is not being forced hidden)...
					if (button:GetAttribute("showbutton")) then
						show = true;
					end
				else
					-- Decrease the gridevent count
					button:SetAttribute("gridevent", max(0, button:GetAttribute("gridevent") - 1));

					-- If we are showing this button (ie. it is not being forced hidden)...
					if (button:GetAttribute("showbutton")) then

						-- If a show grid event is not currently active...
						if (button:GetAttribute("gridevent") == 0) then

							-- Dragging something onto a button may not immediately updated HasAction() so we have to
							-- test for the "receiveddrag" attribute (which we set in the secure code that received
							-- the drag) so that we don't hide the button.

							-- If the button has an action, or if we just dragged something onto this button...
							local actionId = button:GetAttribute("action");
							if ( HasAction(actionId) or actionId == 132 or button:GetAttribute("receiveddrag") ) then  -- 132 == module.controlCancel
								show = true;
							else
								-- The button has no action.
								-- If we want to show empty buttons...
								if (button:GetAttribute("gridshow") > 0) then
									show = true;
								end
							end
						else
							-- There is a show grid event that is currently active.
							-- The user is probably dragging an action button.
							-- Show all buttons, empty or not.
							show = true;
						end
					end
				end
				if (show) then
					button:Show();
				else
					button:Hide();
				end

				button:SetAttribute("receiveddrag", nil);

				-- Next button
				count = count + 1;
				button = group:GetFrameRef("child" .. count);
			end
			-- Next group
			groupNum = groupNum + 1;
			group = self:GetFrameRef("group" .. groupNum);
		end
		
		-- Perform visual updates of the buttons
		self:CallMethod("grideventFunc", showgrid);
	end

	return nil;  -- return a non-false value to allow wrapped script to execute
]=];

-- Create an ActionBarButton that inherits from ActionBarButtonTemplate in ActionBarFrame.xml.
local actButton = CreateFrame("CheckButton", "CT_BarMod_ActionBarButton", UIParent, "ActionBarButtonTemplate");
actButton:ClearAllPoints();
actButton:SetPoint("BOTTOMRIGHT", UIParent, "TOPLEFT", 0, 0);
actButton:SetAttribute("statehidden", true);  -- keep it hidden (Blizzard's code will test for this attribute)
actButton:Hide();

SecureHandlerWrapScript(actButton, "OnAttributeChanged", actButton, actionButton_OnAttributeChanged_Pre_Secure, nil);

actButton.grideventFunc = actionButton_gridevent_Unsecure;

--------------------------------------------
-- Interface

function module:addObjectToGroup(object, groupId)
	local group = getGroup(groupId);
	group:addObject(object);
end

--------------------------------------------
-- Bar Paging

function module:buildPageBasicCondition(groupId)
	-- Build the basic macro conditions needed for bar paging of the specified group.
	local group = getGroup(groupId);
	local bar;
	local condition = "";

	if (groupId == module.actionBarId) then
		condition = condition .. "[bonusbar:5]11; ";
	end

	bar = module:getOption("pageAltKey" .. groupId) or 1;
	if (bar > 1) then
		condition = condition .. "[mod:alt]" .. (bar - 1) .. "; ";
	end

	bar = module:getOption("pageCtrlKey" .. groupId) or 1;
	if (bar > 1) then
		condition = condition .. "[mod:ctrl]" .. (bar - 1) .. "; ";
	end

	bar = module:getOption("pageShiftKey" .. groupId) or 1;
	if (bar > 1) then
		condition = condition .. "[mod:shift]" .. (bar - 1) .. "; ";
	end

	if (groupId == module.actionBarId) then
		for count = 2, 6 do
			condition = condition .. "[bar:" .. count .. "]" .. count .. "; ";
		end
		for count = 1, 4 do
			condition = condition .. "[bonusbar:" .. count .. "]" .. (count + 6) .. "; ";
		end
		condition = condition .. "1";
	else
		condition = condition .. group.num;
	end

	return condition;
end

function module:buildPageAdvancedCondition(groupId)
	-- Build the advanced macro conditions needed for bar paging of the specified group.
	return module.buildCondition(module:getOption("pageCondition" .. groupId) or "");
end

function module:buildPageCondition(groupId)
	-- Build the macro conditions needed for bar paging of the specified group.
	local condition;
	local barPaging = module:getOption("barPaging" .. groupId) or 1;

	if (barPaging == 2) then
		-- Advanced conditions
		condition = module:buildPageAdvancedCondition(groupId);
	else
		-- Basic conditions
		condition = module:buildPageBasicCondition(groupId);
	end

	return condition;
end

function module:registerPagingStateDriver(groupId)
	-- Register the state driver for bar paging.
	if (InCombatLockdown()) then
		module.needRegisterPagingStateDrivers = true;
		return;
	end
	local group = getGroup(groupId);
	RegisterStateDriver(group.frame, "barpage", module:buildPageCondition(groupId));
end

function module:registerAllPagingStateDrivers()
	-- Register all bar paging state drivers.
	if (InCombatLockdown()) then
		return;
	end
	for groupNum, group in pairs(groupList) do
		module:registerPagingStateDriver(group.groupId);
	end
	module.needRegisterPagingStateDrivers = nil;
end
