----------------------------------------------
--               CT_BottomBar                 --
--                                            --
-- Breaks up the main menu bar into pieces,   --
-- allowing you to hide and move the pieces   --
-- independently of each other.               --
-- Please do not modify or otherwise          --
-- redistribute this without the consent of   --
-- the CTMod Team. Thank you.                 --
------------------------------------------------

--------------------------------------------
-- Initialization

local module = { };
local _G = getfenv(0);

local MODULE_NAME = "CT_BottomBar";
local MODULE_VERSION = strmatch(GetAddOnMetadata(MODULE_NAME, "version"), "^([%d.]+)");

module.name = MODULE_NAME;
module.version = MODULE_VERSION;

_G[MODULE_NAME] = module;
CT_Library:registerModule(module);

module.loadedAddons = {};

local appliedOptions;
local pendingOptions;

local frame_SetAlpha;
local frame_EnableMouse;

local savedHideGryphons;

--------------------------------------------
-- Hide Gryphons

local gryphonLoop;

function module:toggleGryphons(hide)
	if (gryphonLoop) then
		-- Prevent infinite loop.
		gryphonLoop = nil;
		return;
	end
	-- Hide/Show the gryphons
	if ( hide ) then
		MainMenuBarLeftEndCap:Hide();
		MainMenuBarRightEndCap:Hide();
	else
		MainMenuBarLeftEndCap:Show();
		MainMenuBarRightEndCap:Show();
	end
	if (type(module.frame) == "table") then
		-- Change the checkbox in CT_BottomBar
		local cb = module.frame.general.hideGryphons;
		cb:SetChecked(hide);
	end
	if (CT_Core) then
		local opt = "hideGryphons";
		if (CT_Core:getOption(opt) ~= hide) then
			-- Change the same option in CT_Core
			gryphonLoop = true;
			CT_Core:setOption(opt, hide, true);
			gryphonLoop = nil;
		end
	end
end

--------------------------------------------
-- Hide Left Texture

local hidLeft;

function module:hideTexturesLeft(hide)
	-- Hide/Show the left textures

	-- MainMenuBarTexture0
	--	- Texture behind action buttons 1 to 6.
	-- MainMenuBarTexture1
	--	- Texture behind action buttons 7 to 12.

	if (hide) then
		MainMenuBarTexture0:Hide();
		MainMenuBarTexture1:Hide();
		hidLeft = true;
	else
		-- Only show the textures if we previously hid them.
		if (not hidLeft) then
			return;
		end
		MainMenuBarTexture0:Show();
		MainMenuBarTexture1:Show();
		hidLeft = false;
	end
end

--------------------------------------------
-- Hide Right Texture

local hidRight;

function module:hideTexturesRight(hide)
	-- Hide/Show the right textures

	-- MainMenuBarTexture2
	--	- Texture behind the action bar arrows, the action bar page number,
	--	  and all but the last two menu micro buttons.
	-- MainMenuBarTexture3
	--	- Texture behind half of the help micro button, the help micro button,
	--	  the key ring button, the bag buttons, and the backpack button.

	if (hide) then
		MainMenuBarTexture2:Hide();
		MainMenuBarTexture3:Hide();
		hidRight = true;
	else
		-- Only show the textures if we previously hid them.
		if (not hidRight) then
			return;
		end
		MainMenuBarTexture2:Show();
		MainMenuBarTexture3:Show();
		hidRight = false;
	end
end

--------------------------------------------
-- Fix CT_BarMod bar positioning issue when CT_BarMod loads before CT_BottomBar.

local function fix_CT_BarMod_BarPositions()

	-- If this is a first run for CT_BarMod...
	if (CT_BarMod and CT_BarMod.isFirstRun and CT_BarMod:isFirstRun()) then

		if (CT_BarMod.resetAllBarPositions) then
			-- Reset the position of all CT_BarMod bars.
			-- This fixes a positioning issue with the bottom left and right
			-- CT_BarMod bars when CT_BarMod loads before CT_BottomBar.
			--
			-- When CT_BarMod starts it can't detect CT_BottomBar since it
			-- isn't loaded yet. It defaults to placing the bottom left and
			-- right action bars above the exp bar. If CT_BottomBar later
			-- loads, those two action bars will then be covering the rep
			-- bar.
			--
			-- Now that CT_BottomBar has loaded, we'll reset the position
			-- of all bars, but only upon detecting a first run of CT_BarMod
			-- since the user won't have had a chance to move the bars
			-- anywhere yet.

			CT_BarMod.resetAllBarPositions();
		end
	end
end

--------------------------------------------
-- Disable action bar

local function hooked_SetAlpha(self, alpha)
	frame_SetAlpha(self, 0);
end

local function hooked_RegisterEvent(self, event)
	self:UnregisterEvent(event);
end

local function hooked_EnableMouse(self, enable)
	if (not InCombatLockdown()) then
		frame_EnableMouse(self, false);
	end
end

local function hooked_Show(self)
	if (not InCombatLockdown()) then
		self:Hide();
	end
end

local function disableActionBar()
	-- Disable the action and bonus action bars.

	-- Disable and hide the ActionButton and BonusActionButton buttons.
	for i = 1, 2 do
		local name, obj;
		if (i == 1) then
			name = "ActionButton";
		else
			name = "BonusActionButton";
		end
		for j = 1, 12 do
			obj = _G[name .. j];
			obj:UnregisterAllEvents();
			obj:EnableMouse(false);
			obj:SetAlpha(0);
			obj:Hide();

			-- Try to keep the alpha at zero.
			hooksecurefunc(obj, "SetAlpha", hooked_SetAlpha);

			-- Try to keep events from being re-registered.
			hooksecurefunc(obj, "RegisterEvent", hooked_RegisterEvent);

			-- Try to keep the mouse disabled.
			hooksecurefunc(obj, "EnableMouse", hooked_EnableMouse);

			-- Try to keep the frame hidden.
			hooksecurefunc(obj, "Show", hooked_Show);
			obj:SetAttribute("statehidden", true);
			RegisterStateDriver(obj, "visibility", "hide");
		end
	end

	-- Because we are continuing to use Blizzard's own buttons for the various
	-- CT_BottomBar bars, we need to keep as much of the default UI active as
	-- possible, so that those bars and their buttons can be updated, shown, and
	-- hidden properly by the default UI.
	--
	-- This means that we can't disable the events on the MainMenuBar,
	-- MainMenuBarArtFrame, or BonusActionBarFrame. We also don't want
	-- to hide these frame. We want Blizzard to continue to manipulate
	-- them.
	--
	-- To keep the BonusActionBarFrame from being seen we will set
	-- its alpha value to zero. Hiding the frame can lead to various
	-- UI elements not being shown/hidden properly.

	local obj;
	obj = BonusActionBarFrame;
	--obj:UnregisterAllEvents();  -- Don't unregister events
	obj:EnableMouse(false);
	obj:SetAlpha(0);

	-- Try to keep the alpha at zero.
	hooksecurefunc(obj, "SetAlpha", hooked_SetAlpha);

	-- Try to keep the mouse disabled.
	hooksecurefunc(obj, "EnableMouse", hooked_EnableMouse);

	-- Don't disable or hide the MainMenuBar. It is used to support
	-- the other bars.

	-- Don't disable or hide the MainMenuBarArtFrame.

	-- Set a flag we can test for to see if the action bar is disabled.
	module.actionBarDisabled = true;
end

--------------------------------------------
-- Configuration

local function convertOldActionBarPosition(oldPosition, oldButtonScale)
	-- Convert the saved position of the old CT_BottomBar action bar drag frame
	-- into a position that can be used to reposition CT_BarMod's action bar frame.
	--
	-- Returns true if conversion worked, else false.
	-- Also returns table containing converted position information if conversion worked.

	if (not CT_BarMod.actionBarId) then
		return false;
	end

	local newDragFrame = _G["CT_BarMod_Group" .. CT_BarMod.actionBarId];
	local newButton = CT_BarModActionButton133;

	if (not newDragFrame or not newButton) then
		return false;
	end

	local oldAncP = oldPosition[1];
	local oldXOffset = oldPosition[4];
	local oldYOffset = oldPosition[5];

	-- The old drag frame.
	-- _G["CT_BottomBarAddon-Action Bar"]
	-- :GetWidth() == 30
	-- :GetHeight() == 1
	local oldDragWidth = 30;
	local oldDragHeight = 1;

	-- The first button that was attached to the old drag frame.
	-- ActionButton1
	-- :GetWidth() == 36
	-- :GetHeight() == 36
	-- :GetPoint() == "BOTTOMLEFT", _G["CT_BottomBarAddon-Action Bar"], "BOTTOMLEFT", 0, 0
	local oldButton = ActionButton1;
	local oldButtonWidth = 36;
	local oldButtonHeight = 36;

	oldButtonWidth = oldButtonWidth * oldButtonScale;
	oldButtonHeight = oldButtonHeight * oldButtonScale;

	-- The new drag frame.
	-- CT_BarMod_Group12
	-- :GetWidth() == 60
	-- :GetHeight() == 30
	local newDragWidth = newDragFrame:GetWidth() or 60;
	local newDragHeight = newDragFrame:GetHeight() or 30;

	-- The first button that is attached to the new drag frame.
	-- CT_BarModActionButton133
	-- :GetWidth() == 36
	-- :GetHeight() == 36
	-- :GetPoint() == "TOPLEFT", CT_BarMod_Group12, "BOTTOMLEFT", 12, -4
	local newButtonScale = oldButtonScale;  -- use same scale for new buttons
	local newButtonWidth = newButton:GetWidth() or 36;
	local newButtonHeight = newButton:GetHeight() or 36;
	local newAncP, newRelF, newRelP, newButtonXOffset, newButtonYOffset = newButton:GetPoint(1);
	newButtonXOffset = abs(newButtonXOffset or 12);
	newButtonYOffset = abs(newButtonYOffset or 4);

	newButtonWidth = newButtonWidth * newButtonScale;
	newButtonHeight = newButtonHeight * newButtonScale;
	newButtonXOffset = newButtonXOffset * newButtonScale;
	newButtonYOffset = newButtonYOffset * newButtonScale;

	local newXOffset;
	local newYOffset;

	if (oldAncP == "TOPLEFT" or oldAncP == "LEFT" or oldAncP == "BOTTOMLEFT") then
		newXOffset = oldXOffset - newButtonXOffset;

	elseif (oldAncP == "TOPRIGHT" or oldAncP == "RIGHT" or oldAncP == "BOTTOMRIGHT") then
		newXOffset = oldXOffset - oldDragWidth + (newDragWidth - newButtonXOffset);

	else -- if (oldAncP == "TOP" or oldAncP == "CENTER" or oldAncP == "BOTTOM") then
		newXOffset = oldXOffset - (oldDragWidth / 2) - newButtonXOffset + (newDragWidth / 2);
	end

	if (oldAncP == "TOPLEFT" or oldAncP == "TOP" or oldAncP == "TOPRIGHT") then
		newYOffset = oldYOffset + newButtonHeight + newDragHeight + newButtonYOffset;

	elseif (oldAncP == "BOTTOMLEFT" or oldAncP == "BOTTOM" or oldAncP == "BOTTOMRIGHT") then
		newYOffset = oldYOffset + newButtonHeight + newButtonYOffset;

	else -- if (oldAncP == "LEFT" or oldAncP == "CENTER" or oldAncP == "RIGHT") then
		newYOffset = oldYOffset + newButtonHeight + newButtonYOffset + (newDragHeight / 2);
	end

	return true, { oldPosition[1], oldPosition[2], oldPosition[3], newXOffset, newYOffset };
end

local function configureActionBar()
	-- Configure/disable the default main action bar.
	-- It may get disabled, CT_BarMod's main action may get enabled and moved, etc.

	local disableMainBar;
	local resetMainBarPos;

	-- Get the current value of the option to disable the default action bar.
	-- This option will only be nil once per character, or per reset of CT_BottomBar options.
	disableMainBar = module:getOption("disableDefaultActionBar");

	if (disableMainBar == nil) then
		-- This is a first run for CT_BottomBar.
		-- (new character, reset of options, or first time using this version)

		-- If CT_BarMod is loaded, and it supports a main action bar...
		if (CT_BarMod and CT_BarMod.actionBarId) then

			-- Disable the default main action bar.
			disableMainBar = true;

			-- Enable the CT_BarMod main action bar
			CT_BarMod:setOption("showGroup" .. CT_BarMod.actionBarId, 1, true);

			-- Reset position of CT_BarMod main action bar.
			resetMainBarPos = true;
		else
			-- CT_BarMod is not loaded, or
			-- This version of CT_BarMod does not support a main action bar.

			-- Do not disable the default main action bar.
			disableMainBar = false;
		end

	else
		-- Since the disableDefaultActionBar option has a value, this is not a first run for CT_BottomBar.

		if (disableMainBar and CT_BarMod and CT_BarMod.actionBarId) then
			-- We are going to disable the main action bar,
			-- and CT_BarMod is loaded,
			-- and this version of CT_BarMod supports a main action bar.

			-- If this is a first run for CT_BarMod...
			if (CT_BarMod.isFirstRun and CT_BarMod:isFirstRun()) then

				-- Enable the CT_BarMod main action bar
				CT_BarMod:setOption("showGroup" .. CT_BarMod.actionBarId, 1, true);

				-- Reset position of CT_BarMod main action bar.
				resetMainBarPos = true;
			end
		end
	end

	-- Possibly disable the game's default action and bonus action bars/buttons.
	-- The default if the option is nil, is true (disable the action bar).
	if ( disableMainBar ) then
		disableActionBar();
	end

	if (resetMainBarPos) then

		-- See if the user has saved data for the action bar from a previous version
		-- of CT_BottomBar.
		local oldPosition = module:getOption("MOVABLE-Action Bar");
		local oldOrientation = module:getOption("orientationAction Bar");
		local oldOpacity = module:getOption("barOpacity");
		local oldScale = module:getOption("barScale");
		local oldSpacing = module:getOption("barSpacing");
		local oldHide = module:getOption("Action Bar");  -- default was not to hide the bar

		local oldData = oldPosition or oldOrientation or oldOpacity or oldScale or oldSpacing or oldHide;

		-- Flag that indicates if we have already converted the old action bar data.
		local oldActionBarConvertedFlag = module:getOption("oldActionBarConvertedFlag");

		-- If we haven't already converted the old action bar data,
		-- and we have old data,
		-- and CT_BarMod is loaded,
		-- and CT_BarMod has the function we need...
		if (not oldActionBarConvertedFlag and oldData and CT_BarMod and CT_BarMod.updateActionBar) then

			-- Try to convert the old CT_BarMod action bar position.
			local converted, newPosition;
			if (oldPosition) then
				converted, newPosition = convertOldActionBarPosition(oldPosition, oldScale or 1);
			end
			if (not converted) then
				-- Will reset to default position instead.
				newPosition = false;
			end

			-- Update CT_BarMod's action bar.
			CT_BarMod.updateActionBar(
				newPosition,
				oldOrientation or "ACROSS",
				oldScale or 1,
				oldOpacity or 1,
				oldSpacing or 6,
				not not oldHide
			);

			-- Set a flag to keep us from using the old data more than once.
			module:setOption("oldActionBarConvertedFlag", true, true);

			-- If this is the first run for CT_BarMod...
			if (CT_BarMod.isFirstRun and CT_BarMod:isFirstRun()) then
				-- When CT_BarMod is first run, it automatically enables group ids 1 through 5,
				-- and disables group ids that are greater than 5.

				-- Disable the CT_BarMod action bars with group ids 1 through 5 (bars 2 through 6).
				for groupId = 1, 5 do
					CT_BarMod:setOption("showGroup" .. groupId, false, true);
				end
			end

		elseif (CT_BarMod and CT_BarMod.resetActionBarPosition) then
			-- Move the main action bar to where the default one is normally located.

			-- We couldn't have CT_BarMod reset the position of the ation bar until
			-- after we disabled the default bar, since CT_BarMod's reset function
			-- checks for a flag which indicates that the default main bar has been
			-- disabled.

			CT_BarMod.resetActionBarPosition();
		end
	end
end

local function configureThings()
	-- Configure the CT_BottomBar bars and environment.

	-- Initialize and configure each addon
	for key, obj in ipairs(module.addons) do
		obj:init();
		if (not obj.isDisabled) then
			obj:setClamped(appliedOptions.clampFrames);
			obj:toggleVisibility(1);
		end
	end

	-- Configure the pet bar using the chosen options.
	local obj = module.ctPetBar;
	if (obj and not obj.isDisabled) then
		module.updatePetBar("petBarScale", appliedOptions.petBarScale);
		module.updatePetBar("petBarOpacity", appliedOptions.petBarOpacity);
		module.updatePetBar("petBarSpacing", appliedOptions.petBarSpacing);
	end

	-- Gryphons
	savedHideGryphons = appliedOptions.hideGryphons;
	module:toggleGryphons(appliedOptions.hideGryphons);

	-- MainMenuBar artwork
	module:hideTexturesLeft(appliedOptions.hideTexturesLeft);
	module:hideTexturesRight(appliedOptions.hideTexturesRight);
end

--------------------------------------------
-- Event handling

local function CT_BottomBar_EventFunc(self, event, arg1, arg2, arg3, ...)

	-- (WoW 3.x notes)
	-- If you reload the UI while in a vehicle, the PLAYER_ENTERING_VEHICLE and PLAYER_ENTERED_VEHICLE
	-- events occur *after* PLAYER_ENTERING_WORLD.
	-- However, if you /exit and click 'exit now', when you restart the game and log back in,
	-- the PLAYER_ENTERING_VEHICLE and PLAYER_ENTERED_VEHICLE events occur *before* PLAYER_ENTERING_WORLD.
	--
	-- (Wow 4.x)
	-- If you log out of the game in a vehicle, when you log back in you won't be in the
	-- vehicle anymore. I'm not sure if the game ejects you from the vehicle before logging out
	-- or when logging back in, or whether this happens for all vehicles or just some.
	-- In 3.x you used to end up back in the vehicle when you returned.

--	print(event, arg1, arg2, arg3, ...);

	if (event == "UNIT_ENTERED_VEHICLE") then
		-- Note: It is possible to enter one vehicle from another (no exited event appears).
		if (arg1 == "player") then
			module.updateBarVisibility();
		end
		module.vehicleSkinName = arg3;

	elseif (event == "UNIT_EXITED_VEHICLE") then
		if (arg1 == "player") then
			module.updateBarVisibility();
		end
		module.vehicleSkinName = nil;

	elseif (event == "PLAYER_ENTERING_WORLD") then
		module.updateBarVisibility();
		module.updateVehicleVisibility();

	elseif (event == "PLAYER_REGEN_DISABLED") then
		-- About to enter combat lockdown.
		module:applyPendingOptions();

		module.updateBarVisibility(true);
		module.updateVehicleVisibility();

	elseif (event == "PLAYER_REGEN_ENABLED") then
		-- Combat lockdown is now over.
		module:applyPendingOptions();

		module.updateBarVisibility();
		module.updateVehicleVisibility();

	elseif (event == "PLAYER_LOGIN") then
		-- Gryphons
		--
		-- When CT_BottomBar and CT_Core are both loaded, we want CT_BottomBar's
		-- gryphon option to have priority over the CT_Core one when the addons
		-- are loading.
		--
		-- CT_Core's default is to not hide the gryphons. CT_BottomBar's default
		-- is to hide the gryphons.
		--
		-- Since CT_BottomBar loads before CT_Core, CT_BottomBar will apply its
		-- gryphon option first, but since CT_Core is not loaded CT_BottomBar
		-- cannot update CT_Core's option.
		--
		-- When CT_Core loads, it applies its gryphon option and updates the
		-- CT_BottomBar option since both addons are now loaded. We've saved
		-- the current CT_BottomBar option in a variable that we will now
		-- apply at PLAYER_LOGIN time so that we can override CT_Core's option.
		module:toggleGryphons(savedHideGryphons);

	end
end

local function registerEvents()
	-- Rregister events.

	local frame = CT_BottomBar_Frame;

--	frame:RegisterEvent("UNIT_ENTERING_VEHICLE");
--	frame:RegisterEvent("UNIT_EXITING_VEHICLE");

	frame:RegisterEvent("UNIT_ENTERED_VEHICLE");
	frame:RegisterEvent("UNIT_EXITED_VEHICLE");

	frame:RegisterEvent("PLAYER_REGEN_ENABLED");
	frame:RegisterEvent("PLAYER_REGEN_DISABLED");

	frame:RegisterEvent("PLAYER_ENTERING_WORLD");
	frame:RegisterEvent("PLAYER_LOGIN");

	frame:SetScript("OnEvent", CT_BottomBar_EventFunc);
end

--------------------------------------------
-- The main menu bar's vehicle leave button

local function CT_BottomBar_MainMenuBarVehicleLeaveButton_Update()
	-- This is a modified version of MainMenuBarVehicleLeaveButton_Update in MainMenuBar.lua.
	--
	-- The game shows a vehicle leave button when the vehicle that the player
	-- is in does not have its own vehicle UI. This is a different button than
	-- the one the game displays on the vehicle menu bar when the player has
	-- a vehicle UI.

	-- This function checks the same conditions that Blizzard does,
	-- and in the same sequence.
	--
	-- We will hide the button if Blizzard has attached it to one
	-- of our enabled bars, otherwise we'll leave it attached and
	-- visible on Blizzard's bar.

	if ( CanExitVehicle() ) then
		-- MainMenuBarVehicleLeaveButton:ClearAllPoints();
		local obj;
		if ( IsPossessBarVisible() ) then
			obj = module.ctPossess;
		elseif ( GetNumShapeshiftForms() > 0 ) then
			obj = module.ctClassBar;
		elseif ( HasMultiCastActionBar() ) then
			obj = module.ctMultiCast;
		else
			obj = module.ctPossess;
		end
		if ( obj and not obj.isDisabled ) then
			MainMenuBarVehicleLeaveButton:Hide();
		end
	end
end

--------------------------------------------
-- UIParent_ManageFramePositions

local function CT_BottomBar_Hooked_UIParent_ManageFramePositions()
	-- (hooksecurefunc of UIParent_ManageFramePosition in UIParent.lua)
	-- This func and the hooks used to be at the end of the CT_BottomBar_ShapeshiftBar_Init function.

	local obj;

	obj = module.ctPetBar;
	if (obj) then
		if (not obj.isDisabled) then
			-- Hide pet bar textures
			obj:toggleVisibility(3, false);
		else
			-- Update saved info about the textures.
			obj:saveTextures();
		end
	end

	obj = module.ctClassBar;
	if (obj) then
		if (not obj.isDisabled) then
			-- Hide class bar textures
			obj:toggleVisibility(3, false);
		else
			-- Update saved info about the textures.
			obj:saveTextures();
		end
	end
end

--------------------------------------------
-- Hook functions

local function hook_UIParent_ManageFramePositions()
	-- Hook the UIParent_ManageFramePositions function so we can make some adjustments.

	-- This hook catches the places where the function is called using its global name.
	hooksecurefunc("UIParent_ManageFramePositions", CT_BottomBar_Hooked_UIParent_ManageFramePositions);

	-- The following hooks are required for Blizzard xml scripts that use the following syntax to
	-- call UIParent_ManageFramePositions. This xml syntax does not call our secure hook of
	-- UIParent_ManageFramePositions, so we have to explicitly hook anything that calls it to
	-- ensure our function gets called.
	--
	-- 	<OnShow function="UIParent_ManageFramePositions"/>
	-- 	<OnHide function="UIParent_ManageFramePositions"/>

	-- From BonusActionBarFrame.xml
	ShapeshiftBarFrame:HookScript("OnShow", CT_BottomBar_Hooked_UIParent_ManageFramePositions);
	ShapeshiftBarFrame:HookScript("OnHide", CT_BottomBar_Hooked_UIParent_ManageFramePositions);

	-- From BonusActionBarFrame.xml
	PossessBarFrame:HookScript("OnShow", CT_BottomBar_Hooked_UIParent_ManageFramePositions);
	PossessBarFrame:HookScript("OnHide", CT_BottomBar_Hooked_UIParent_ManageFramePositions);

	-- From DurabilityFrame.xml
	DurabilityFrame:HookScript("OnShow", CT_BottomBar_Hooked_UIParent_ManageFramePositions);
	DurabilityFrame:HookScript("OnHide", CT_BottomBar_Hooked_UIParent_ManageFramePositions);

	-- From MainMenuBar.xml
	MainMenuBarMaxLevelBar:HookScript("OnShow", CT_BottomBar_Hooked_UIParent_ManageFramePositions);
	MainMenuBarMaxLevelBar:HookScript("OnHide", CT_BottomBar_Hooked_UIParent_ManageFramePositions);

	-- From MultiCastActionBarFrame.xml
	MultiCastActionBarFrame:HookScript("OnShow", CT_BottomBar_Hooked_UIParent_ManageFramePositions);
	MultiCastActionBarFrame:HookScript("OnHide", CT_BottomBar_Hooked_UIParent_ManageFramePositions);

	-- From PetActionBarFrame.xml
	PetActionBarFrame:HookScript("OnShow", CT_BottomBar_Hooked_UIParent_ManageFramePositions);
	PetActionBarFrame:HookScript("OnHide", CT_BottomBar_Hooked_UIParent_ManageFramePositions);

	-- From ReputationFrame.xml
	ReputationWatchBar:HookScript("OnHide", CT_BottomBar_Hooked_UIParent_ManageFramePositions);
end

local function hookFunctions()
	-- Hook various Blizzard functions.

	-- Hook the function Blizzard uses to position various bars on the screen, hide/show some textures, etc.
	hook_UIParent_ManageFramePositions();

	-- Hook the OnShow script of the MainMenuBarVehicleLeaveButton.
	-- The only time Blizzard shows this button is during MainMenuBarVehicleLeaveButton_Update
	-- which is in MainMenuBar.lua. They show it right after they've positioned the button, and
	-- before they call ShowPetActionBar(true). We may want to hide the button before the pet
	-- bar function gets called since the pet bar's position may be calculated using the :GetRight()
	-- value of the leave button if the button is shown.
	MainMenuBarVehicleLeaveButton:HookScript("OnShow", CT_BottomBar_MainMenuBarVehicleLeaveButton_Update);
end

--------------------------------------------
-- CT_BottomBar initialization

module.update = function(self, optName, value)
	-- optName -- "init == Perform initialization.
	-- optName -- name of option to be changed.

	if (optName ~= "init") then
		-- Update an option.
		return module:updateOption(optName, value);
	end

	-- Initialize CT_BottomBar.
	-- This will be the first code run by CT_Library once the game has loaded CT_BottomBar.

	-- Create a frame for the addon (used for events, etc).
	local frame = CreateFrame("Frame", "CT_BottomBar_Frame");
	frame_SetAlpha = frame.SetAlpha;
	frame_EnableMouse = frame.EnableMouse;

	-- Fix CT_BarMod bar positioning issue when CT_BarMod loads before CT_BottomBar.
	fix_CT_BarMod_BarPositions()

	-- Configure/disable the default main action bar.
	configureActionBar();

	-- Disable mouse in MainMenuBar
	MainMenuBar:EnableMouse(false);

	-- Make sure Blizzard managed frames are in position.
	UIParent_ManageFramePositions();

	-- Create the options tables.
	module.appliedOptions = {};
	module.pendingOptions = {};

	appliedOptions = module.appliedOptions;
	pendingOptions = module.pendingOptions;

	-- Initialize some other lua files.
	module:optionsInit();
	module:addonsInit();

	-- Load the addons (bars).
	do
		-- We want to do this before initializing the applied options table.
		--
		-- "Experience Bar" needs to load before "Reputation Bar".
		--
		module:loadAddon("Action Bar Arrows");
		module:loadAddon("Bags Bar");
		module:loadAddon("Key Ring");  -- Show after bags bar in options window
		module:loadAddon("Class Bar");
		module:loadAddon("Experience Bar");
		module:loadAddon("Reputation Bar");  -- Show after exp bar in options window
		module:loadAddon("Menu Bar");
		module:loadAddon("Pet Bar");
		module:loadAddon("Possess Bar");
		module:loadAddon("MultiCastBar");  -- Totem bar
		module:loadAddon("Vehicle Bar");
	end

	-- Hook some functions.
	hookFunctions();

	-- Register some events.
	registerEvents();

	-- Initialize the applied options table.
	module:optionsInitApplied();

	-- Configure things
	configureThings();

	-- Update the "disable the default action bar" option.
	-- This will also update CT_BarMod's version of this option.
	module:setOption("disableDefaultActionBar", module.actionBarDisabled, true);
end

--------------------------------------------
-- Slash command.

local function slashCommand(msg)
	module:showModuleOptions(module.name);
end

module:setSlashCmd(slashCommand, "/ctbb", "/ctbottom", "/ctbottombar");
