------------------------------------------------
--               CT_BottomBar                 --
--                                            --
-- Breaks up the main menu bar into pieces,   --
-- allowing you to hide and move the pieces   --
-- independently of each other.               --
-- Please do not modify or otherwise          --
-- redistribute this without the consent of   --
-- the CTMod Team. Thank you.                 --
------------------------------------------------

--------------------------------------------
-- Initialization

local _G = getfenv(0);
local module = _G.CT_BottomBar;

local ctRelativeFrame = module.ctRelativeFrame;
local appliedOptions;

local setpoint;
local setparent;

--------------------------------------------
-- Experience Bar

local function addon_SetWidth(self, width)
	width = width or EXP_DEFAULT_WIDTH or 1024;
	local exhaustionStateID, exhaustionStateName, exhaustionStateMultiplier = GetRestState();
	if (exhaustionStateID ~= nil) then  -- nil test needed to prevent error in ExhaustionTick_OnEvent when logging in.
		MainMenuExpBar_SetWidth(width);
	end
end

local function addon_UpdateNoVehicle(self)
	-- Update the experience bar when player does not have a vehicle ui
	-- self == experience bar object

	local objExp = MainMenuExpBar;
	local objMax = MainMenuBarMaxLevelBar;

	objExp:ClearAllPoints();
	setparent(objExp, self.frame);
	objExp:EnableMouse(true);
	setpoint(objExp, "BOTTOMLEFT", self.frame, 0, 0);
	objExp:SetFrameLevel(objExp:GetParent():GetFrameLevel());

	addon_SetWidth(self, EXP_DEFAULT_WIDTH);

	objMax:ClearAllPoints();
	setparent(objMax, self.frame);
--	setpoint(objMax, "TOPLEFT", objExp);
	setpoint(objMax, "TOPLEFT", objExp, "TOPLEFT", 0, -11);

	addon_SetWidth(self, EXP_DEFAULT_WIDTH);
end

local function addon_UpdateInVehicle(self)
	-- Update the experience bar frame when in a vehicle
	-- self == exp bar object

	local skinName = module.vehicleSkinName;
	if (not skinName) then
		return;
	end

	local expBar = MainMenuExpBar;
	local vehBar = VehicleMenuBar;

	-- The numbers and points are from VehicleMenuBar_MoveMicroButtons() in VehicleMenuBar.lua.

	expBar:ClearAllPoints();			
	setparent(expBar, vehBar);
	expBar:EnableMouse(true);

	if (skinName == "Natural") then
		setpoint(expBar, "TOPLEFT", vehBar, "TOPLEFT", 166, 53);
		setpoint(expBar, "TOPRIGHT", vehBar, "TOPRIGHT", -166, 53);

		addon_SetWidth(self, vehBar:GetWidth() - 166 - 166);

		expBar:SetFrameLevel(expBar:GetParent():GetFrameLevel());

	else -- if (skinName == "Mechanical") then
		setpoint(expBar, "TOPLEFT", vehBar, "TOPLEFT", 148, 50);
		setpoint(expBar, "TOPRIGHT", vehBar, "TOPRIGHT", -150, 50);

		addon_SetWidth(self, vehBar:GetWidth() - 148 - 150);

		expBar:SetFrameLevel(expBar:GetParent():GetFrameLevel());		
	end

	return;
end

local function addon_UpdateBonusBar(self)
	-- Update the experience bar frame when on a special bonus bar frame.
	-- self == exp bar object

	local barType = GetBonusBarOverrideBarType() or "default";
	if (barType == "default") then
		return;
	end

	local expBar = MainMenuExpBar;
	local parent = BonusActionBarFrame;
	local barInfo = BonusActionBarGetBarInfo(barType);

	-- The numbers and points are from BonusActionBarFrame.lua.

	expBar:ClearAllPoints();			
	setparent(expBar, parent);
	expBar:EnableMouse(true);

	addon_SetWidth(self, barInfo.xpWidth);

	setpoint(expBar, "TOP", parent, 0, 0);

	return;
end

local function addon_Update(self)
	-- Update the experience bar frame
	-- self == exp bar object

	local hideVehicle = appliedOptions.vehicleHideFrame;
	if  UnitHasVehicleUI("player") then
		if (not hideVehicle) then
			-- Position the frames on the vehicle.
			addon_UpdateInVehicle(self);
			return;
		end
	elseif (not module.actionBarDisabled) then
		local barType = GetBonusBarOverrideBarType() or "default";
		if (barType ~= "default") then
			addon_UpdateBonusBar(self);
			return;
		end
	end

	-- Frames are not on a vehicle
	-- Position the frames where we want them.

	local objExp = MainMenuExpBar;
	local objMax = MainMenuBarMaxLevelBar;

	self.helperFrame:ClearAllPoints();
	self.helperFrame:SetPoint("TOPLEFT", objExp, 0, 0);
	self.helperFrame:SetPoint("BOTTOMRIGHT", objExp);

	addon_UpdateNoVehicle(self);
end

local function addon_Hooked_SetPoint(...)
	-- (hooksecurefunc of each bar's SetPoint method)
	if (module.ctExpBar.isDisabled) then
		return;
	end
	addon_Update(module.ctExpBar);
end

local function addon_Hooked_SetParent(self, parent)
	-- (hooksecurefunc of each bar's SetParent method)
	if (module.ctExpBar.isDisabled) then

		-- If the default action bar is disabled...
		if (module.actionBarDisabled) then

			-- Only do this if the exp bar and the default action bar
			-- have both been disabled.

			-- If the game tries to parent the exp bar to the
			-- BonusActionBarFrame, then it means that the new (WoW 4.0)
			-- larger bonus action bar frame is being displayed.
			--
			-- Because the default action bar has been disabled the
			-- BonusActionBarFrame should not be visible.
			--
			-- We want to disable the mouse in the exp bar to prevent
			-- the mouse from interfering with anything above the exp bar
			-- while it is parented to the BonusActionBarFrame.

			if (parent == BonusActionBarFrame) then
				self:EnableMouse(false);
			else
				self:EnableMouse(true);
			end
		end

		return;
	end
	addon_Update(module.ctExpBar);
end

local function addon_Disable(self)
	if (UnitHasVehicleUI("player")) then
		addon_UpdateInVehicle(self);

	elseif (not module.actionBarDisabled) then
		local barType = GetBonusBarOverrideBarType() or "default";
		if (barType ~= "default") then
			addon_UpdateBonusBar(self);
		end
	end
end

local function addon_Init(self)
	-- Initialization
	-- self == exp bar object

	appliedOptions = module.appliedOptions;

	module.ctExpBar = self;
	module.CT_BottomBar_ExpBar_Update = addon_Update;

	local frame = CreateFrame("Frame", "CT_BottomBar_" .. self.frameName .. "_GuideFrame");
	self.helperFrame = frame;

	setpoint = frame.SetPoint;
	setparent = frame.SetParent;

	return true;
end

local function addon_PostInit(self)
	-- Initialization
	-- self == object

	local frames = self.frames;

	for i = 2, 3 do
		hooksecurefunc(frames[i], "SetPoint", addon_Hooked_SetPoint);
		hooksecurefunc(frames[i], "SetParent", addon_Hooked_SetParent);
	end
end

local function addon_Register()
	module:registerAddon(
		"Experience Bar",  -- option name
		"ExpBar",  -- used in frame names
		"Experience Bar",  -- shown in options window & tooltips
		"Experience Bar",  -- title for horizontal orientation
		nil,  -- title for vertical orientation
		{ "BOTTOMLEFT", ctRelativeFrame, "BOTTOM", -512, 42 },
		{ -- settings
			orientation = "ACROSS",
			usedOnVehicle = true,
			usedOnBonusBar = true,
		},
		addon_Init,
		addon_PostInit,
		addon_Update,
		nil,  -- no orientation function
		nil,  -- no enable function
		addon_Disable,
		"helperFrame",
		MainMenuExpBar,
		MainMenuBarMaxLevelBar,
		ExhaustionTick
	);
end

module.loadedAddons["Experience Bar"] = addon_Register;
